#define DICT "abcdefghijklmnopqrstuvwxyz"
#define DICTLENGTH 26

typedef unsigned int uint32;
typedef unsigned long long int uint64;
typedef unsigned char uchar8;

typedef struct RainbowConfig 
{
    uint32 minPlainLen; /* Minimum password length      */
    uint32 maxPlainLen; /* Maximum password length      */
    uint32 chainLen;    /* Length of each chain         */
    uint32 tableLen;    /* Number of chains per table   */
    uint32 algorithm;   /* Algorithm Index to use       */
    uint32 dictLen;     /* Length of the dictionary     */
    uint32 tableIdx;    /* The table index to use       */
    uint32 startIdx;    /* The start index to use       */
} RainbowConfig;

__constant__ __align__(16) RainbowConfig config;
__constant__  uchar8 dictionary[256];;
__constant__  uint64 iRange[16];

__device__ int indexToPlain(uint64 index, char *plain);
__device__ uint64 hashToIndex(uint64 hash, int pos);

#ifndef HASH_FUNCTION
#define HASH_FUNCTION

#define	F(b,c,d)	((((c) ^ (d)) & (b)) ^ (d))
#define G(b,c,d)	(((b) & (c)) | ((b) & (d)) | ((c) & (d)))
#define	H(b,c,d)	((b) ^ (c) ^ (d))

#define ROTATE(a,n)     (((a)<<(n))|(((a)&0xffffffff)>>(32-(n))))

#define R0(a,b,c,d,k,s,t) { \
	a+=((k)+(t)+F((b),(c),(d))); \
	a=ROTATE(a,s); };

#define R1(a,b,c,d,k,s,t) { \
	a+=((k)+(t)+G((b),(c),(d))); \
	a=ROTATE(a,s); };\

#define R2(a,b,c,d,k,s,t) { \
	a+=((k)+(t)+H((b),(c),(d))); \
	a=ROTATE(a,s); };

__noinline__ __device__ uint64 plainToHash(uint32 x0, uint32 x1, uint32 x2, uint32 size)
{

    uint32 A, B, C, D;
    uint32 X[16];

    A = 0x67452301; B = 0xEFCDAB89; C = 0x98BADCFE;  D = 0x10325476;

    X[0]  = x0; 
    X[1]  = x1; 
    X[2]  = x2; 
    X[3]  = 0;
    X[4]  = 0; 
    X[5]  = 0; 
    X[6]  = 0; 
    X[7]  = 0;
    X[8]  = 0; 
    X[9]  = 0; 
    X[10] = 0; 
    X[11] = 0;
    X[12] = 0; 
    X[13] = 0; 
    X[14] = size<<3; 
    X[15] = 0;

    ((uchar8 *)X)[size] = 0x80;




	R0(A,B,C,D,X[ 0], 3,0); R0(D,A,B,C,X[ 1], 7,0);	
    R0(C,D,A,B,X[ 2],11,0);	R0(B,C,D,A,X[ 3],19,0);	
    R0(A,B,C,D,X[ 4], 3,0);	R0(D,A,B,C,X[ 5], 7,0);	
    R0(C,D,A,B,X[ 6],11,0);	R0(B,C,D,A,X[ 7],19,0);	
    R0(A,B,C,D,X[ 8], 3,0);	R0(D,A,B,C,X[ 9], 7,0);	
    R0(C,D,A,B,X[10],11,0);	R0(B,C,D,A,X[11],19,0);	
	R0(A,B,C,D,X[12], 3,0);	R0(D,A,B,C,X[13], 7,0); 
    R0(C,D,A,B,X[14],11,0); R0(B,C,D,A,X[15],19,0);

	R1(A,B,C,D,X[ 0], 3,0x5A827999); R1(D,A,B,C,X[ 4], 5,0x5A827999);
	R1(C,D,A,B,X[ 8], 9,0x5A827999); R1(B,C,D,A,X[12],13,0x5A827999);
	R1(A,B,C,D,X[ 1], 3,0x5A827999); R1(D,A,B,C,X[ 5], 5,0x5A827999);
	R1(C,D,A,B,X[ 9], 9,0x5A827999); R1(B,C,D,A,X[13],13,0x5A827999);
	R1(A,B,C,D,X[ 2], 3,0x5A827999); R1(D,A,B,C,X[ 6], 5,0x5A827999);
	R1(C,D,A,B,X[10], 9,0x5A827999); R1(B,C,D,A,X[14],13,0x5A827999);
	R1(A,B,C,D,X[ 3], 3,0x5A827999); R1(D,A,B,C,X[ 7], 5,0x5A827999);
	R1(C,D,A,B,X[11], 9,0x5A827999); R1(B,C,D,A,X[15],13,0x5A827999);

	R2(A,B,C,D,X[ 0], 3,0x6ED9EBA1); R2(D,A,B,C,X[ 8], 9,0x6ED9EBA1);
	R2(C,D,A,B,X[ 4],11,0x6ED9EBA1); R2(B,C,D,A,X[12],15,0x6ED9EBA1);
	R2(A,B,C,D,X[ 2], 3,0x6ED9EBA1); R2(D,A,B,C,X[10], 9,0x6ED9EBA1);
	R2(C,D,A,B,X[ 6],11,0x6ED9EBA1); R2(B,C,D,A,X[14],15,0x6ED9EBA1);
	R2(A,B,C,D,X[ 1], 3,0x6ED9EBA1); R2(D,A,B,C,X[ 9], 9,0x6ED9EBA1);
	R2(C,D,A,B,X[ 5],11,0x6ED9EBA1); R2(B,C,D,A,X[13],15,0x6ED9EBA1);
	R2(A,B,C,D,X[ 3], 3,0x6ED9EBA1); R2(D,A,B,C,X[11], 9,0x6ED9EBA1);
	R2(C,D,A,B,X[ 7],11,0x6ED9EBA1); R2(B,C,D,A,X[15],15,0x6ED9EBA1);

    A += 0x67452301; B += 0xEFCDAB89; C += 0x98BADCFE; D += 0x10325476;

    return ((uint64)B <<32) | A ;

}

#endif



extern "C"    
__global__ void start(  uint64 * results )
{

    uint32 i;
    uint64 index;
    uint32 length;
    uint32 plain[3];

    /* This is the starting point for each chain */
    index = threadIdx.x + (blockIdx.x<<8);


    for(i = 0; i < config.chainLen - 1 ; i++)
    {
        /* reset the plaintext */

        plain[0] = 0;
        plain[1] = 0;
        plain[2] = 0;
        
        length = indexToPlain(index, (char*)plain);
        __syncthreads();

        index = plainToHash( plain[0], plain[1], plain[2], length);

        index = hashToIndex(index, i);





    }

    *results  = index;


} 

__device__ int indexToPlain(uint64 index, char *plain)
{
    int i;
    uint64 pLength;

    for (i = config.maxPlainLen - 1; i >= config.minPlainLen - 1; i--)
	{
        
		if (index >= iRange[i])
		{
			pLength = i + 1;
			break;
		}
	}



    index -= iRange[pLength-1];

    for (i = 0; i < pLength; i++)
	{
        plain[i] = dictionary[index % DICTLENGTH]; 

		index /= DICTLENGTH;
	}

    return pLength;

}


__device__ uint64 hashToIndex(uint64 hash, int pos)
{

    return (hash + config.tableIdx + pos) % iRange[config.maxPlainLen];


}